from enum import StrEnum

from pydantic import BaseModel, Field, model_serializer, model_validator


class Icon(BaseModel):
    """Icon model for use in the application.

    Example:
        `Icon(
            name="pen-fill",
            snippet="<i class='bi bi-pen-fill'></i>",
            toolkit="bootstrap"
        )`
    """

    name: str = Field(description="The name of the icon.")
    snippet: str = Field(default="", description="The HTML snippet for the icon.")
    svg: str = Field(default="", description="The SVG snippet for the icon.")
    toolkit: str | None = Field(default=None, description="The toolkit for the icon.")

    def __str__(self) -> str:
        return self.name

    @model_validator(mode="after")
    def check_snippet_or_svg(self):
        assert not (self.snippet == "" and self.svg == "")
        return self

    @model_serializer
    def serialize_model(self):
        if self.has_svg:
            return {"name": self.name, "snippet": self.svg}
        return {"name": self.name, "snippet": self.snippet}

    @property
    def has_snippet(self) -> bool:
        return self.snippet is not None

    @property
    def has_svg(self) -> bool:
        return self.svg is not None

    @property
    def html(self):
        return self.svg if self.has_svg else self.snippet


PENCIL = Icon(
    name="pen-fill", snippet="<i class='bi bi-pen-fill'></i>", toolkit="bootstrap"
)
CLOCK_HISTORY = Icon(
    name="clock-history",
    snippet='<i class="bi bi-clock-history"></i>',
    toolkit="bootstrap",
)
ARROWS_COLLAPSE = Icon(
    name="arrows-collapse",
    snippet='<i class="bi bi-arrows-collapse"></i>',
    toolkit="bootstrap",
)


class LibraryIcon(StrEnum):
    """HTML referencing icon.

    Must include library if required (e.g. for bootstrap/font-awesome/etc.)


    Example include:
    ```
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap-icons@1.11.3/font/bootstrap-icons.min.css" />
    ```

    Attributes:
        PENCIL: Edit icon.
        TRASH_CAN: Delete icon.
        PLUS_CIRCLE: Create icon.
        ENVELOPE_OPEN: Detail icon.
        LIST_UL: List icon.
        DATABASE_FILL_GEAR: Admin icon.
        CLOCK_HISTORY: History icon.
        ARROWS_COLLAPSE: Expand icon.
        ARROW_CLOCKWISE: Refresh icon.
        CHECK: Check icon.
        X: X icon.
        BOX_ARROW_UP: Check out icon.
        BOX_ARROW_IN_DOWN: Check in icon.
        UPDATE: Edit icon.
        ADMIN: Admin icon.
        DELETE: Delete icon.
        CREATE: Create icon.
        DETAIL: Detail icon.
        LIST: List icon.
        HISTORY: History icon.
        CHECKOUT: Check out icon.
        CHECKIN: Check in icon.
    """  # noqa: E501

    PENCIL = PENCIL.snippet
    TRASH_CAN = "<i class='bi bi-trash2'></i>"
    PLUS_CIRCLE = '<i class="bi bi-plus-circle"></i>'
    ENVELOPE_OPEN = '<i class="bi bi-envelope-open"></i>'
    LIST_UL = '<i class="bi bi-list-ul"></i>'
    DATABASE_FILL_GEAR = '<i class="bi bi-database-fill-gear"></i>'
    CLOCK_HISTORY = CLOCK_HISTORY.snippet
    ARROWS_COLLAPSE = ARROWS_COLLAPSE.snippet
    ARROW_CLOCKWISE = '<i class="bi bi-arrow-clockwise"></i>'
    CHECK = '<i class="bi bi-check2-square">'
    X = '<i class="bi bi-x"></i>'
    BOX_ARROW_UP = '<i class="bi bi-box-arrow-up"></i>'
    BOX_ARROW_IN_DOWN = '<i class="bi bi-box-arrow-in-down"></i>'

    UPDATE = PENCIL
    ADMIN = DATABASE_FILL_GEAR
    DELETE = TRASH_CAN
    CREATE = PLUS_CIRCLE
    DETAIL = ENVELOPE_OPEN
    LIST = LIST_UL
    HISTORY = CLOCK_HISTORY
    CHECKOUT = BOX_ARROW_UP
    CHECKIN = BOX_ARROW_IN_DOWN



def sort_title(title: str) -> str:
    """Sort a title by the first word, ignoring articles.

    Articles include "a", "an", and "the".

    Examples:
        >>> sort_title("The Cat in the Hat")
        'Cat in the Hat, The'
        >>> sort_title("A Tale of Two Cities")
        'Tale of Two Cities, A'
        >>> sort_title("My Fair Lady")
        'My Fair Lady'

    Args:
        title: The title to be sorted.

    Returns:
        The sorted title.
    """
    articles = {"a", "an", "the"}

    title = title.lower()

    first, _, rest = title.partition(" ")
    return f"{rest}, {first}" if first in articles else title


def truncate_secret(
    secret: str, *, max_length: int, mask: str = "*", mask_short: bool = False
) -> str:
    """Truncate a secret to a maximum length and mask truncated characters.

    The secret is truncated to the specified length by removing characters from the
    middle of the path.

    Examples:
        >>> truncate_secret("i72BPzV54LH7lwaez5F5BF9gRuvX5Phy", max_length=20, mask=".")
        'i72BPzV...9gRuvX5Phy'
        >>> truncate_secret("C:/Users/username/Documents/file.txt", max_length=30)
        'i72BPzV54LH7***5F5BF9gRuvX5Phy'
        >>> truncate_secret(
            "i72BPzV54LH7lwaez5F5BF9gRuvX5Phy",
            max_length=40,
            mask_short=True
        )
        '****************************************'
        >>> truncate_secret(
            "i72B",
            max_length=8,
            mask_short=True
        )
        '********'

    Args:
        secret (str): The secret to be truncated.
        max_length (int): The maximum length of the truncated string.
        mask (str): The character to use for masking the truncated characters.
        mask_short (bool): Whether to mask the secret if it is already shorter than the
            maximum length.

    Returns:
        str: The truncated string.

    Raises:
        ValueError: If the secret is already shorter than the maximum length and
            mask_short is False.
    """

    if len(secret) <= max_length and not mask_short:
        msg = f"Secret is already shorter than max_length [{max_length}]"
        raise ValueError(msg)
    if len(secret) <= max_length:
        return mask * max_length

    tail_length = max_length // 2
    head_length = max_length - tail_length - 3

    head = secret[:head_length]
    tail = secret[-tail_length:]

    new_string = f"{head}{3 * mask}{tail}"

    assert len(new_string) <= max_length

    return new_string



"""Widgets for Django models."""

from django.db.models.fields import Field
from django.template.loader import render_to_string
from pydantic import BaseModel


class DetailWidget(BaseModel):
    field_class: type[Field]
    template_name: str

    def render_to_string(self, list_or_queryset):
        tag_dict = {"values": list_or_queryset}
        return render_to_string(self.template_name, context=tag_dict)



from typing import cast

from django.db.models import Model
from django.db.models.fields import Field


def get_model_name(model: type[Model]) -> str:
    """Get model's name as defined in Meta class.

    Will return as all lower case

    Args:
        model: the model to get name from

    Returns:
        model's name
    """
    return model._meta.model_name  # noqa: SLF001


def get_model_verbose_name_plural(model: type[Model]) -> str:
    """Get model's verbose name as defined in Meta class.

    Args:
        model: the model to get verbose name from

    Returns:
        model's verbose name
    """
    return model._meta.verbose_name_plural  # noqa: SLF001


def get_model_label(model: Model) -> str:
    """Get model's label as defined in Meta class.

    Args:
        model: the model to get label from

    Returns:
        model's label"""
    return model._meta.label  # noqa: SLF001


def get_model_fields(
    model: type[Model],
    *,
    fields: list[str] | None = None,
    exclude: list[str] | None = None,
) -> list[Field]:
    """Returns list of fields from model.

    Cannot include both fields and exclude.

    Args:
      model: the model to get fields from
      fields: list of fields to include, default is None
      exclude: list of fields to exclude, default is None

    Returns:
      list of fields
    """
    if fields is not None and exclude is not None:
        msg = "Cannot specify both 'fields' and 'exclude'."
        raise ValueError(msg)

    field_list_full = cast(list[Field], list(model._meta.fields))  # noqa: SLF001  # `_meta` is private

    if fields is not None:
        return [model._meta.get_field(field) for field in fields]  # noqa: SLF001  # `_meta` is private
    if exclude is not None:
        return [field for field in field_list_full if field.name not in exclude]

    return field_list_full



"""Useful functions for default values in Django models."""

from datetime import date, datetime

from django.utils import timezone


def django_today() -> date:
    """Return the current date in the timezone of the Django settings.

    Example:
        >>> from django_rubble.utils.default_funcs import django_today
        >>> django_today()
        datetime.date(2021, 1, 1)
    """
    return timezone.now().date()


def django_now() -> datetime:
    """Return the current datetime in the timezone of the Django settings.

    Example:
    >>> from django_rubble.utils.default_funcs import django_now
    >>> django_now()
    datetime.datetime(2021, 1, 1, 0, 0, tzinfo=<UTC>)
    """
    return timezone.now()



from factory import Faker
from factory.django import DjangoModelFactory

from django_rubble.models.number_models import TestNumberedModel


class NumberedModelFactory(DjangoModelFactory):
    class Meta:
        model = TestNumberedModel

    description = Faker("sentence", nb_words=4)
